@@ -1,33 +1,37 @@
-# -*- coding: iso-8859-1 -*-
+# -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2005 Edgewall Software
+# Copyright (C) 2004-2006 Edgewall Software
 # Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
+# Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution. The terms
-# are also available at http://trac.edgewall.com/license.html.
+# are also available at http://trac.edgewall.org/wiki/TracLicense.
 #
 # This software consists of voluntary contributions made by many
 # individuals. For the exact contribution history, see the revision
-# history and logs, available at http://projects.edgewall.com/trac/.
+# history and logs, available at http://trac.edgewall.org/log/.
 #
 # Author: Christopher Lenz <cmlenz@gmx.de>
 
-from __future__ import generators
 import re
+from StringIO import StringIO
 import time
 
 from trac.core import *
+from trac.db import get_column_names
 from trac.perm import IPermissionRequestor
 from trac.ticket import Ticket, TicketSystem
-from trac.util import escape, unescape, format_datetime, http_date, \
-                      shorten_line, CRLF, Markup
+from trac.util.datefmt import format_datetime, http_date
+from trac.util.html import escape, html, unescape
+from trac.util.text import shorten_line, CRLF
 from trac.web import IRequestHandler
-from trac.web.chrome import add_link, add_stylesheet, INavigationContributor
-from trac.wiki import wiki_to_html, wiki_to_oneliner, IWikiMacroProvider, \
-                      IWikiSyntaxProvider
-
+from trac.web.chrome import add_link, add_script, add_stylesheet, \
+                            INavigationContributor
+from trac.wiki import wiki_to_html, wiki_to_oneliner, IWikiSyntaxProvider
+from trac.wiki.macros import WikiMacroBase
+from trac.mimeview.api import Mimeview, IContentConverter
 
 class QuerySyntaxError(Exception):
     """Exception raised when a ticket query cannot be parsed from a string."""
@@ -57,6 +61,8 @@ class Query(object):
 
     def from_string(cls, env, string, **kw):
         filters = string.split('&')
+        kw_strs = ['order', 'group']
+        kw_bools = ['desc', 'groupdesc', 'verbose']
         constraints = {}
         for filter in filters:
             filter = filter.split('=')
@@ -75,7 +81,16 @@ class Query(object):
                 neg = '!'
                 field = field[:-1]
             values = map(lambda x: neg + mode + x, values)
-            constraints[field] = values
+            try:
+                field = str(field)
+                if field in kw_strs:
+                    kw[field] = values[0]
+                elif field in kw_bools:
+                    kw[field] = True
+                else:
+                    constraints[field] = values
+            except UnicodeError:
+                pass # field must be a str, see `get_href()`
         return cls(env, constraints, **kw)
     from_string = classmethod(from_string)
 
@@ -130,7 +145,7 @@ class Query(object):
 
         return self.cols
 
-    def execute(self, db=None):
+    def execute(self, req, db=None):
         if not self.cols:
             self.get_columns()
 
@@ -141,13 +156,13 @@ class Query(object):
             db = self.env.get_db_cnx()
         cursor = db.cursor()
         cursor.execute(sql, args)
-        columns = cursor.description
+        columns = get_column_names(cursor)
         results = []
         for row in cursor:
             id = int(row[0])
-            result = {'id': id, 'href': self.env.href.ticket(id)}
+            result = {'id': id, 'href': req.href.ticket(id)}
             for i in range(1, len(columns)):
-                name, val = columns[i][0], row[i]
+                name, val = columns[i], row[i]
                 if name == self.group:
                     val = val or 'None'
                 elif name == 'reporter':
@@ -161,14 +176,17 @@ class Query(object):
         cursor.close()
         return results
 
-    def get_href(self, format=None):
-        return self.env.href.query(order=self.order,
-                                   desc=self.desc and 1 or None,
-                                   group=self.group or None,
-                                   groupdesc=self.groupdesc and 1 or None,
-                                   verbose=self.verbose and 1 or None,
-                                   format=format,
-                                   **self.constraints)
+    def get_href(self, req, order=None, desc=None, format=None):
+        # FIXME: only use .href from that 'req' for now
+        if desc is None:
+            desc = self.desc
+        if order is None:
+            order = self.order
+        return req.href.query(order=order, desc=desc and 1 or None,
+                              group=self.group or None,
+                              groupdesc=self.groupdesc and 1 or None,
+                              verbose=self.verbose and 1 or None,
+                              format=format, **self.constraints)
 
     def get_sql(self):
         """Return a (sql, params) tuple for the query."""
@@ -329,7 +347,25 @@ class Query(object):
 
 class QueryModule(Component):
 
-    implements(IRequestHandler, INavigationContributor, IWikiSyntaxProvider)
+    implements(IRequestHandler, INavigationContributor, IWikiSyntaxProvider,
+               IContentConverter)
+
+    # IContentConverter methods
+    def get_supported_conversions(self):
+        yield ('rss', 'RSS Feed', 'xml',
+               'trac.ticket.Query', 'application/rss+xml', 8)
+        yield ('csv', 'Comma-delimited Text', 'csv',
+               'trac.ticket.Query', 'text/csv', 8)
+        yield ('tab', 'Tab-delimited Text', 'tsv',
+               'trac.ticket.Query', 'text/tab-separated-values', 8)
+
+    def convert_content(self, req, mimetype, query, key):
+        if key == 'rss':
+            return self.export_rss(req, query)
+        elif key == 'csv':
+            return self.export_csv(req, query, mimetype='text/csv')
+        elif key == 'tab':
+            return self.export_csv(req, query, '\t', 'text/tab-separated-values')
 
     # INavigationContributor methods
 
@@ -339,9 +375,9 @@ class QueryModule(Component):
     def get_navigation_items(self, req):
         from trac.ticket.report import ReportModule
         if req.perm.has_permission('TICKET_VIEW') and \
-           not self.env.is_component_enabled(ReportModule):
-            yield 'mainnav', 'tickets', Markup('<a href="%s">View Tickets</a>',
-                                               self.env.href.query())
+                not self.env.is_component_enabled(ReportModule):
+            yield ('mainnav', 'tickets',
+                   html.A('View Tickets', href=req.href.query()))
 
     # IRequestHandler methods
 
@@ -375,14 +411,14 @@ class QueryModule(Component):
             for var in ('query_constraints', 'query_time', 'query_tickets'):
                 if req.session.has_key(var):
                     del req.session[var]
-            req.redirect(query.get_href())
+            req.redirect(query.get_href(req))
 
-        add_link(req, 'alternate', query.get_href('rss'), 'RSS Feed',
-                 'application/rss+xml', 'rss')
-        add_link(req, 'alternate', query.get_href('csv'),
-                 'Comma-delimited Text', 'text/plain')
-        add_link(req, 'alternate', query.get_href('tab'), 'Tab-delimited Text',
-                 'text/plain')
+        # Add registered converters
+        for conversion in Mimeview(self.env).get_supported_conversions(
+                                             'trac.ticket.Query'):
+            add_link(req, 'alternate',
+                     query.get_href(req, format=conversion[0]),
+                     conversion[1], conversion[3])
 
         constraints = {}
         for k, v in query.constraints.items():
@@ -400,16 +436,12 @@ class QueryModule(Component):
         req.hdf['query.constraints'] = constraints
 
         format = req.args.get('format')
-        if format == 'rss':
-            self.display_rss(req, query)
-            return 'query_rss.cs', 'application/rss+xml'
-        elif format == 'csv':
-            self.display_csv(req, query)
-        elif format == 'tab':
-            self.display_csv(req, query, '\t')
-        else:
-            self.display_html(req, query)
-            return 'query.cs', None
+        if format:
+            Mimeview(self.env).send_converted(req, 'trac.ticket.Query', query,
+                                              format, 'query')
+
+        self.display_html(req, query)
+        return 'query.cs', None
 
     # Internal methods
 
@@ -434,7 +466,6 @@ class QueryModule(Component):
             vals = req.args[field]
             if not isinstance(vals, (list, tuple)):
                 vals = [vals]
-            vals = map(lambda x: x.value, vals)
             if vals:
                 mode = req.args.get(field + '_mode')
                 if mode:
@@ -470,6 +501,7 @@ class QueryModule(Component):
     def display_html(self, req, query):
         req.hdf['title'] = 'Custom Query'
         add_stylesheet(req, 'common/css/report.css')
+        add_script(req, 'common/js/query.js')
 
         db = self.env.get_db_cnx()
 
@@ -493,14 +525,19 @@ class QueryModule(Component):
                 req.hdf['query.constraints.%s.values.%d' % (field, idx)] = ''
 
         cols = query.get_columns()
-        for i in range(len(cols)):
-            header = {'name': cols[i]}
-            req.hdf['query.headers.%d' % i] = header
+        labels = dict([(f['name'], f['label']) for f in query.fields])
+        for idx, col in enumerate(cols):
+            req.hdf['query.headers.%d' % idx] = {
+                'name': col, 'label': labels.get(col, 'Ticket'),
+                'href': query.get_href(req, order=col,
+                                       desc=(col == query.order and
+                                             not query.desc))
+            }
 
-        href = self.env.href.query(group=query.group,
-                                   groupdesc=query.groupdesc and 1 or None,
-                                   verbose=query.verbose and 1 or None,
-                                   **query.constraints)
+        href = req.href.query(group=query.group,
+                              groupdesc=query.groupdesc and 1 or None,
+                              verbose=query.verbose and 1 or None,
+                              **query.constraints)
         req.hdf['query.order'] = query.order
         req.hdf['query.href'] = href
         if query.desc:
@@ -512,22 +549,24 @@ class QueryModule(Component):
         if query.verbose:
             req.hdf['query.verbose'] = True
 
-        tickets = query.execute(db)
+        tickets = query.execute(req, db)
         req.hdf['query.num_matches'] = len(tickets)
 
         # The most recent query is stored in the user session
         orig_list = rest_list = None
         orig_time = int(time.time())
-        if str(query.constraints) != req.session.get('query_constraints'):
-            # New query, initialize session vars
-            req.session['query_constraints'] = str(query.constraints)
-            req.session['query_time'] = int(time.time())
+        query_constraints = unicode(query.constraints)
+        if query_constraints != req.session.get('query_constraints') \
+                or int(req.session.get('query_time', 0)) < orig_time - 3600:
+            # New or outdated query, (re-)initialize session vars
+            req.session['query_constraints'] = query_constraints
             req.session['query_tickets'] = ' '.join([str(t['id']) for t in tickets])
         else:
             orig_list = [int(id) for id in req.session.get('query_tickets', '').split()]
             rest_list = orig_list[:]
             orig_time = int(req.session.get('query_time', 0))
-        req.session['query_href'] = query.get_href()
+        req.session['query_href'] = query.get_href(req)
+        req.session['query_time'] = orig_time
 
         # Find out which tickets originally in the query results no longer
         # match the constraints
@@ -535,13 +574,18 @@ class QueryModule(Component):
             for tid in [t['id'] for t in tickets if t['id'] in rest_list]:
                 rest_list.remove(tid)
             for rest_id in rest_list:
-                ticket = Ticket(self.env, int(rest_id), db=db)
-                data = {'id': ticket.id, 'time': ticket.time_created,
-                        'changetime': ticket.time_changed, 'removed': True,
-                        'href': self.env.href.ticket(ticket.id)}
-                data.update(ticket.values)
+                try:
+                    ticket = Ticket(self.env, int(rest_id), db=db)
+                    data = {'id': ticket.id, 'time': ticket.time_created,
+                            'changetime': ticket.time_changed, 'removed': True,
+                            'href': req.href.ticket(ticket.id)}
+                    data.update(ticket.values)
+                except TracError, e:
+                    data = {'id': rest_id, 'time': 0, 'changetime': 0,
+                            'summary': html.EM(e)}
                 tickets.insert(orig_list.index(rest_id), data)
 
+        num_matches_group = {}
         for ticket in tickets:
             if orig_list:
                 # Mark tickets added or changed since the query was first
@@ -551,6 +595,8 @@ class QueryModule(Component):
                 elif int(ticket['changetime']) > orig_time:
                     ticket['changed'] = True
             for field, value in ticket.items():
+                if field == query.group:
+                    num_matches_group[value] = num_matches_group.get(value, 0)+1
                 if field == 'time':
                     ticket[field] = format_datetime(value)
                 elif field == 'description':
@@ -559,6 +605,7 @@ class QueryModule(Component):
                     ticket[field] = value
 
         req.hdf['query.results'] = tickets
+        req.hdf['query.num_matches_group'] = num_matches_group
         req.session['query_tickets'] = ' '.join([str(t['id']) for t in tickets])
 
         # Kludge: only show link to available reports if the report module is
@@ -566,39 +613,42 @@ class QueryModule(Component):
         from trac.ticket.report import ReportModule
         if req.perm.has_permission('REPORT_VIEW') and \
            self.env.is_component_enabled(ReportModule):
-            req.hdf['query.report_href'] = self.env.href.report()
-
-    def display_csv(self, req, query, sep=','):
-        req.send_response(200)
-        req.send_header('Content-Type', 'text/plain;charset=utf-8')
-        req.end_headers()
+            req.hdf['query.report_href'] = req.href.report()
 
+    def export_csv(self, req, query, sep=',', mimetype='text/plain'):
+        content = StringIO()
         cols = query.get_columns()
-        req.write(sep.join([col for col in cols]) + CRLF)
+        content.write(sep.join([col for col in cols]) + CRLF)
 
-        results = query.execute(self.env.get_db_cnx())
+        results = query.execute(req, self.env.get_db_cnx())
         for result in results:
-            req.write(sep.join([str(result[col]).replace(sep, '_')
-                                                .replace('\n', ' ')
-                                                .replace('\r', ' ')
-                                for col in cols]) + CRLF)
+            content.write(sep.join([unicode(result[col]).replace(sep, '_')
+                                                        .replace('\n', ' ')
+                                                        .replace('\r', ' ')
+                                    for col in cols]) + CRLF)
+        return (content.getvalue(), '%s;charset=utf-8' % mimetype)
 
-    def display_rss(self, req, query):
+    def export_rss(self, req, query):
         query.verbose = True
         db = self.env.get_db_cnx()
-        results = query.execute(db)
+        results = query.execute(req, db)
         for result in results:
-            result['href'] = self.env.abs_href.ticket(result['id'])
+            result['href'] = req.abs_href.ticket(result['id'])
             if result['reporter'].find('@') == -1:
                 result['reporter'] = ''
             if result['description']:
-                # str() cancels out the Markup() returned by wiki_to_html
-                result['description'] = str(wiki_to_html(result['description'] or '',
-                                                         self.env, req, db,
-                                                         absurls=1))
+                # unicode() cancels out the Markup() returned by wiki_to_html
+                descr = wiki_to_html(result['description'], self.env, req, db,
+                                     absurls=True)
+                result['description'] = unicode(descr)
             if result['time']:
                 result['time'] = http_date(result['time'])
         req.hdf['query.results'] = results
+        req.hdf['query.href'] = req.abs_href.query(group=query.group,
+                groupdesc=query.groupdesc and 1 or None,
+                verbose=query.verbose and 1 or None,
+                **query.constraints)
+        return (req.hdf.render('query_rss.cs'), 'application/rss+xml')
 
     # IWikiSyntaxProvider methods
     
@@ -609,77 +659,66 @@ class QueryModule(Component):
         yield ('query', self._format_link)
 
     def _format_link(self, formatter, ns, query, label):
-        if query[0] == '?':
-            return '<a class="query" href="%s">%s</a>' \
-                   % (escape(formatter.href.query() + query.replace(' ', '+')),
-                      label)
+        if query.startswith('?'):
+            return html.A(label, class_='query',
+                          href=formatter.href.query() + query.replace(' ', '+'))
         else:
-            from trac.ticket.query import Query, QuerySyntaxError
             try:
                 query = Query.from_string(formatter.env, query)
-                return '<a class="query" href="%s">%s</a>' \
-                       % (escape(query.get_href()), label)
+                return html.A(label, href=query.get_href(formatter), # Hack
+                              class_='query')
             except QuerySyntaxError, e:
-                return '<em class="error">[Error: %s]</em>' % escape(e)
+                return html.EM('[Error: %s]' % e, class_='error')
 
 
-class QueryWikiMacro(Component):
+class TicketQueryMacro(WikiMacroBase):
     """Macro that lists tickets that match certain criteria.
     
     This macro accepts two parameters, the second of which is optional.
-
+    
     The first parameter is the query itself, and uses the same syntax as for
-    {{{query:}}} wiki links. The second parameter determines how the list of
-    tickets is presented: the default presentation is to list the ticket ID next
-    to the summary, with each ticket on a separate line. If the second parameter
-    is given and set to '''compact''' then the tickets are presented as a
-    comma-separated list of ticket IDs.
+    `query:` wiki links (but '''not''' the variant syntax starting with "?").
+
+    The second parameter determines how the list of tickets is presented:
+    the default presentation is to list the ticket ID next to the summary,
+    with each ticket on a separate line.
+    If the second parameter is given, it must be one of:
+     - '''compact''' -- the tickets are presented as a comma-separated
+       list of ticket IDs. 
+     - '''count''' -- only the count of matching tickets is displayed
     """
-    implements(IWikiMacroProvider)
-
-    def get_macros(self):
-        yield 'TicketQuery'
-
-    def get_macro_description(self, name):
-        import inspect
-        return inspect.getdoc(QueryWikiMacro)
 
     def render_macro(self, req, name, content):
         query_string = ''
-        compact = 0
+        compact = False
+        count = False
         argv = content.split(',')
         if len(argv) > 0:
             query_string = argv[0]
             if len(argv) > 1:
-                if argv[1].strip().lower() == 'compact':
-                    compact = 1
-        
-        try:
-            from cStringIO import StringIO
-        except NameError:
-            from StringIO import StringIO
-        buf = StringIO()
+                format = argv[1].strip().lower()
+                if format == 'compact':
+                    compact = True
+                elif format == 'count':
+                    count = True
 
         query = Query.from_string(self.env, query_string)
         query.order = 'id'
-        tickets = query.execute()
+        tickets = query.execute(req)
         if tickets:
+            def ticket_anchor(ticket):
+                return html.A('#%s' % ticket['id'],
+                              class_=ticket['status'],
+                              href=req.href.ticket(int(ticket['id'])),
+                              title=shorten_line(ticket['summary']))
             if compact:
-                links = []
-                for ticket in tickets:
-                    href = self.env.href.ticket(int(ticket['id']))
-                    summary = escape(shorten_line(ticket['summary']))
-                    links.append('<a class="%s ticket" href="%s" '
-                                 'title="%s">#%s</a>' % (ticket['status'], href,
-                                 summary, ticket['id']))
-                buf.write(', '.join(links))
+                alist = [ticket_anchor(ticket) for ticket in tickets]
+                return html.SPAN(alist[0], *[(', ', a) for a in alist[1:]])
+            elif count:
+                cnt = len(tickets)
+                return html.SPAN(cnt, title='%d tickets for which %s' % 
+                                 (cnt, query_string))
             else:
-                buf.write('<dl class="wiki compact">')
-                for ticket in tickets:
-                    href = self.env.href.ticket(int(ticket['id']))
-                    buf.write('<dt><a href="%s">#%s</a></dt>' % (href,
-                                                                 ticket['id']))
-                    buf.write('<dd>%s</dd>' % (escape(ticket['summary'])))
-                buf.write('</dl>')
-
-        return buf.getvalue()
+                return html.DL([(html.DT(ticket_anchor(ticket)),
+                                 html.DD(ticket['summary']))
+                                for ticket in tickets], class_='wiki compact')
